#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

char *ttymod="mod_ttygdilite.dll";
char *cpumod="M65C02/mod_M65C02.dll";

typedef struct
{
 HMODULE  dll;
 uint32_t *mod_type;
 uint8_t  **mod_name;
 uint8_t  **mod_pref;
 uint8_t  **mod_vers;
 uint8_t  **mod_copr;
 void    (*write16)   (uint16_t a, uint8_t v);
 void    (*write256)  (uint16_t a, uint8_t v);
 void    (*write2048) (uint16_t a, uint8_t v);
 uint8_t (*read16)    (uint16_t a);
 uint8_t (*read256)   (uint16_t a);
 uint8_t (*read2048)  (uint16_t a);
 int     (*init)      (void);
 int     (*deinit)    (void);
} slot;

/* will eventually need to be increased to 131072 and 32768 */
uint8_t  ram[65536], rom[12288];

uint8_t  fnstr[256];

/* Data pointers and functinos for TTY driver */
uint32_t *tty_mod_type;
uint8_t  **tty_mod_name;
uint8_t  **tty_mod_pref;
uint8_t  **tty_mod_vers;
uint8_t  **tty_mod_copr;
uint8_t  *ttykey;
void     (*ttywrite)   (uint16_t a, uint8_t v);
uint8_t  (*ttyread)    (uint16_t a);
void     (*ttyout)     (uint8_t a, uint8_t v);
uint8_t  (*ttyin)      (uint8_t a);
void     (*ttydeinit)  (void);
void     (*ttyinit)    (void);
int      (*ttyloop)    (void);

/* Data pointers and functinos for CPU driver */
uint32_t *cpu_mod_type;
uint8_t  **cpu_mod_name;
uint8_t  **cpu_mod_pref;
uint8_t  **cpu_mod_vers;
uint8_t  **cpu_mod_copr;
void     (*cpureset)   (void);
void     (*cpurun)     (void);
void     (*cpuinit)    (void (*v_cpuwrite)(uint16_t, uint8_t), uint8_t (*v_cpuread)(uint16_t), int8_t (*v_cpuloop)(void));
void     (*cpuseta)    (uint8_t val);
void     (*cpusetx)    (uint8_t val);
void     (*cpusety)    (uint8_t val);
void     (*cpusets)    (uint8_t val);
void     (*cpusetp)    (uint8_t val);
void     (*cpusetpc)   (uint16_t val);
void     (*cpusetrate) (int val);
uint8_t  (*cpugeta)    (void);
uint8_t  (*cpugetx)    (void);
uint8_t  (*cpugety)    (void);
uint8_t  (*cpugets)    (void);
uint8_t  (*cpugetp)    (void);
uint16_t (*cpugetpc)   (void);
int      (*cpugetrate) (void);

HMODULE mod_tty, mod_cpu;

void notamod (char *header, char *filename)
{
 if (mod_tty) FreeLibrary(mod_tty);
 if (mod_cpu) FreeLibrary(mod_cpu);
 fprintf (stderr, "%s: %s is not a MoDapple module.\n", header, filename);
 exit(1);
}

void nofunc (char *header, char *filename, char *ident)
{
 if (mod_tty) FreeLibrary(mod_tty);
 if (mod_cpu) FreeLibrary(mod_cpu);
 fprintf (stderr, "%s: Could not find a '%s' function in %s.\n", header, ident, filename);
 exit(1);
}

#define TFUNCTINO(a,b) sprintf (fnstr,"%s_%s",*tty_mod_pref,b); if (!(a=(void*)GetProcAddress(mod_tty,fnstr))) nofunc(progname, ttymod, b)
#define CFUNCTINO(a,b) sprintf (fnstr,"%s_%s",*cpu_mod_pref,b); if (!(a=(void*)GetProcAddress(mod_cpu,fnstr))) nofunc(progname, cpumod, b)

char *progname="MoDapple";

int bload (char *filename, int expect, uint8_t *pointer)
{
 FILE *file;
 long l;
 
 file=fopen(filename,"rb");
 if (!file)
 {
  fprintf (stderr, "%s: ", progname);
  perror(filename);
  return -1;
 }
 
 fseek(file,0,SEEK_END);
 l=ftell(file);
 fseek(file,0,SEEK_SET);
 if (l!=expect)
 {
  fclose(file);
  fprintf (stderr, "%s: %s: Wrong file size (%lu expected, %lu found)\n", progname, filename, expect, l);
  return -1;
 }
 l=fread(pointer,1,expect,file);
 fclose(file);
 if (l<expect)
 {
  fprintf(stderr, "%s: %s: Short read (%lu expected, %lu found)\n", progname, filename, expect, l);
  return -1;
 }
 
 printf ("%s: Opened binary blob %s (%lu bytes)\n", progname, filename, expect);
 return 0;
}

void cpuwrite(uint16_t a, uint8_t v)
{
 if ((a>1023)&&(a<3072))  {ttywrite(a,v); return;}
 if ((a>8191)&&(a<24576)) {ttywrite(a,v); return;}
 if (a<49152)             {ram[a]=v; return;}
 
 if (a>=0xD000)
 {
  return;
 }
 
 if ((a&0xFFF0)==0xC000)  {ttyout(a,v); return;}
 if ((a&0xFFF0)==0xC010)  {ttyout(a,v); return;}
 if ((a&0xFFF0)==0xC030)  {ttyout(a,v); return;}
 if ((a&0xFFF0)==0xC050)  {ttyout(a,v); return;}
}

uint8_t cpuread(uint16_t a)
{
 if ((a>1023)&&(a<3072))  return ttyread(a);
 if ((a>8191)&&(a<24576)) return ttyread(a);
 if (a<49152)             return ram[a];
 
 if (a>=0xD000)
 {
  return rom[a-0xD000];
 }
 
 if ((a&0xFFF0)==0xC000)  return ttyin(a);
 if ((a&0xFFF0)==0xC010)  return ttyin(a);
 if ((a&0xFFF0)==0xC030)  return ttyin(a);
 if ((a&0xFFF0)==0xC050)  return ttyin(a);
 
 return 0;
}

static int die;

int8_t cpuloop(void)
{
 int x;
 
 sprintf (fnstr, "%04X- A=%02X X=%02X Y=%02X S=%02X P=%02X",
          cpugetpc(), cpugeta(), cpugetx(), cpugety(), cpugets(), cpugetp());
 //SetConsoleTitle(fnstr);
 
 x=ttyloop();
 if (x==-1) return -1;
 if (x==1) cpureset();
 return 0;
}

int WINAPI WinMain(HINSTANCE ghinstance, HINSTANCE dummy, LPSTR cmdline, int nCmdShow)
{
 mod_tty=0;
 mod_cpu=0;
 
 mod_tty=LoadLibrary(ttymod);
 if (!mod_tty)
 {
  fprintf(stderr, "%s: Could not open module %s\n", progname, ttymod);
  return -1;
 }
 printf ("%s: Opened module %s\n", progname, ttymod);
 if (!(tty_mod_type=(void*)GetProcAddress(mod_tty,"mod_type"))) notamod(progname, ttymod);
 if (!(tty_mod_name=(void*)GetProcAddress(mod_tty,"mod_name"))) notamod(progname, ttymod);
 if (!(tty_mod_pref=(void*)GetProcAddress(mod_tty,"mod_pref"))) notamod(progname, ttymod);
 if (!(tty_mod_vers=(void*)GetProcAddress(mod_tty,"mod_vers"))) notamod(progname, ttymod);
 if (!(tty_mod_copr=(void*)GetProcAddress(mod_tty,"mod_copr"))) notamod(progname, ttymod);
 
 printf ("Type: %c%c%c%c\n", ((char*)tty_mod_type)[0], ((char*)tty_mod_type)[1], 
                             ((char*)tty_mod_type)[2], ((char*)tty_mod_type)[3]);

 if ((*tty_mod_type)!=0x50595454)
 {
  FreeLibrary(mod_tty);
  fprintf(stderr, "%s: %s is not a MoDapple console module.\n", progname, ttymod);
  return -1;
 }
 
 printf ("%s, Version %s\n%s\n", *tty_mod_name, *tty_mod_vers, *tty_mod_copr);
 
 TFUNCTINO(ttykey,"ttykey");
 TFUNCTINO(ttywrite,"ttywrite");
 TFUNCTINO(ttyread,"ttyread");
 TFUNCTINO(ttyout,"ttyout");
 TFUNCTINO(ttyin,"ttyin");
 TFUNCTINO(ttydeinit,"ttydeinit");
 TFUNCTINO(ttyinit,"ttyinit");
 TFUNCTINO(ttyloop,"ttyloop");
 
 printf ("Link successful!\n");
 
 mod_cpu=LoadLibrary(cpumod);
 if (!mod_cpu)
 {
  fprintf(stderr, "%s: Could not open module %s\n", progname, cpumod);
  return -1;
 }
 printf ("%s: Opened module %s\n", progname, cpumod);
 if (!(cpu_mod_type=(void*)GetProcAddress(mod_cpu,"mod_type"))) notamod(progname, cpumod);
 if (!(cpu_mod_name=(void*)GetProcAddress(mod_cpu,"mod_name"))) notamod(progname, cpumod);
 if (!(cpu_mod_pref=(void*)GetProcAddress(mod_cpu,"mod_pref"))) notamod(progname, cpumod);
 if (!(cpu_mod_vers=(void*)GetProcAddress(mod_cpu,"mod_vers"))) notamod(progname, cpumod);
 if (!(cpu_mod_copr=(void*)GetProcAddress(mod_cpu,"mod_copr"))) notamod(progname, cpumod);
 
 printf ("Type: %c%c%c%c\n", ((char*)cpu_mod_type)[0], ((char*)cpu_mod_type)[1], 
                             ((char*)cpu_mod_type)[2], ((char*)cpu_mod_type)[3]);

 if ((*cpu_mod_type)!=0x32303536)
 {
  FreeLibrary(mod_tty);
  FreeLibrary(mod_cpu);
  fprintf(stderr, "%s: %s is not a MoDapple CPU module.\n", progname, cpumod);
  return -1;
 }
 
 printf ("%s, Version %s\n%s\n", *cpu_mod_name, *cpu_mod_vers, *cpu_mod_copr);

 CFUNCTINO(cpureset,"cpureset");
 CFUNCTINO(cpurun,"cpurun");
 CFUNCTINO(cpuinit,"cpuinit");
 CFUNCTINO(cpuseta,"cpuseta");
 CFUNCTINO(cpusetx,"cpusetx");
 CFUNCTINO(cpusety,"cpusety");
 CFUNCTINO(cpusets,"cpusets");
 CFUNCTINO(cpusetp,"cpusetp");
 CFUNCTINO(cpusetpc,"cpusetpc");
 CFUNCTINO(cpusetrate,"cpusetrate");
 CFUNCTINO(cpugeta,"cpugeta");
 CFUNCTINO(cpugetx,"cpugetx");
 CFUNCTINO(cpugety,"cpugety");
 CFUNCTINO(cpugets,"cpugets");
 CFUNCTINO(cpugetp,"cpugetp");
 CFUNCTINO(cpugetpc,"cpugetpc");
 CFUNCTINO(cpugetrate,"cpugetrate");

 printf ("Link successful!\n");
 
 if (bload("apple2p.rom",12288,rom)) return -1;
 
 cpuinit(cpuwrite, cpuread, cpuloop);

 cpureset();
 cpusetrate(65);
 
 printf ("CPU start address: %04X\n", cpugetpc());
 
 *ttykey=nCmdShow;
 ttyinit();
 cpurun();
 ttydeinit();
 
 FreeLibrary(mod_tty);
 return 0;
}
